<?php

namespace Drupal\field_states\Form;

use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Serialization\Yaml;

/**
 * Transition form.
 */
class StateTransitionForm extends FormBase implements StateTransitionFormInterface {

  /**
   * The entity.
   *
   * @var \Drupal\Core\Entity\ContentEntityInterface
   */
  protected $entity;

  /**
   * The state field name.
   *
   * @var string
   */
  protected $fieldName;

  /**
   * {@inheritdoc}
   */
  public function getEntity() {
    return $this->entity;
  }

  /**
   * {@inheritdoc}
   */
  public function setEntity(ContentEntityInterface $entity) {
    $this->entity = $entity;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getFieldName() {
    return $this->fieldName;
  }

  /**
   * {@inheritdoc}
   */
  public function setFieldName($field_name) {
    $this->fieldName = $field_name;
    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getBaseFormId() {
    return 'state_machine_transition_form';
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    $entity = $this->getEntity();
    if (!$entity) {
      throw new \RuntimeException('No entity provided to StateTransitionForm.');
    }
    // Example ID: "state_machine_transition_form_commerce_order_state_1".
    $form_id = $this->getBaseFormId();
    $form_id .= '_' . $entity->getEntityTypeId() . '_' . $this->fieldName;
    $form_id .= '_' . $entity->id();

    return $form_id;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    /** @var \Drupal\state_machine\Plugin\Field\FieldType\StateItemInterface $state_item */
    $state_item = $this->entity->get($this->fieldName)->first();
    $state = !empty($state_item) ? $state_item->get('value')->getString() : '';
    $form['actions'] = [
      '#type' => 'container',
    ];
    $transitions = [];
    if (!empty($state_item)) {
      $transitions = $state_item->getPossibleTransitions($state);
    }
    else {
      $workflow = $this->getWorkflow();
      foreach ($workflow['transitions'] as $transition) {
        if (in_array($state, $transition['from'])) {
          $transitions[] = $transition;
        }
      }
    }
    foreach ($transitions as $transition_id => $transition) {
      $transition['id'] = $transition_id;
      $form['actions'][$transition_id] = [
        '#type' => 'submit',
        '#value' => $transition['label'],
        '#submit' => ['::submitForm'],
        '#transition' => $transition,
      ];
    }
    if (empty($transitions)) {
      $form['#empty_transition'] = TRUE;
      $form['#state'] = $state;
    }

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $triggering_element = $form_state->getTriggeringElement();
    /** @var \Drupal\state_machine\Plugin\Workflow\WorkflowTransition $transition */
    $transition = $triggering_element['#transition'];
    /** @var \Drupal\state_machine\Plugin\Field\FieldType\StateItemInterface $state_item */
    $state_item = $this->entity->get($this->fieldName)->first();
    // Ensure the transition is still allowed before applying it.
    $transition['state'] = $this->entity->get($this->fieldName)->value;
    $isAllowed = !empty($state_item) && $state_item->isTransitionAllowed($transition);
    if ($isAllowed) {
      $state_item->applyTransition($transition);
    }
    else {
      $this->entity->get($this->fieldName)
        ->setValue(['value' => $transition['to']]);
      if (!empty($transition['workflow'])) {
        $type = \Drupal::service('plugin.manager.states.workflow');
        $workflows = is_array($transition['workflow']) ? $transition['workflow'] : [$transition['workflow']];
        foreach ($workflows as $workflowId) {
          $plugin = $type->createInstance($workflowId);
          if (!empty($plugin) && method_exists($plugin, 'action')) {
            $plugin->action($transition, $this->getWorkflow(), $this->entity);
          }
        }
      }
    }
    $this->entity->save();
    if ($isAllowed) {
      $state_item->applyAction($transition);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getWorkflow() {
    $fieldName = $this->fieldName;
    $fieldDefinition = $this->getEntity()->getFieldDefinition($fieldName);
    $workflow = trim($fieldDefinition->getSetting('workflow'));
    if (!empty($workflow)) {
      $workflow = Yaml::decode($workflow);
    }
    return $workflow;
  }

}
